!******************************************************************************
!
! File : anderson_realmix.f90  -- (modified version of anderson_mixing.f90)
!   by : Alan Tackett
!   on : 07/19/99
!  for : density mixing (originally written for pwpaw code)
!
!  This module contains routines to implement the extended Anderson
!  mixing method as outlined in V. Eyert, J. Comp. Phys. 124,  271(1996).
!
!  The method is defined by eq's: 2.1, 8.2, 7.7
!
!******************************************************************************

Module anderson_realmix

Implicit NONE!!!!!!
Save

Type Anderson_Context  !** Anderson Mixing context
    Real*8    :: NewMix    !** Amount of new vectors to mix, ie beta in paper.
    Integer :: Nmax      !** Max number of vectors to keep 
    Integer :: N         !** Current number of vectors in list
    Integer :: Slot      !** New fill Slot
    Integer :: VecSize   !** Size of each vector
    Integer :: Err_Unit  !** Error unit
    Real*8, Pointer :: Matrix(:,:)
    Real*8, Pointer :: Gamma(:)  !** Gamma as defined in 7.6
    Real*8, Pointer :: DF(:,:)   !** Delta F
    Real*8, Pointer :: Fprev(:)
    Real*8, Pointer :: DX(:,:)
    Real*8, Pointer :: Xprev(:)

! temporary constants and arrays needed for each call to Anderson_Mix
    Integer, Pointer :: IPIV(:)
    Real*8,  Pointer :: S(:)
    Real*8,  Pointer :: RWork(:)
    Real*8, Pointer :: U(:,:)
    Real*8, Pointer :: VT(:,:)
    Real*8, Pointer :: Work(:)
    Real*8, Pointer :: DupMatrix(:,:)
    Integer          :: Lwork
    Integer          :: LRwork
    Real*8           :: ConditionNo
    Real*8           :: MachAccur
End Type
    
!******************************************************************************
Contains
!******************************************************************************


!******************************************************************************
!
! Anderson_Mix - Performs the actual mixing of the input vector with the
!                history and retuns the result.
!
!   AC - Anderson context
!   X  - Current vector on input and new guess on output
!   F  - F(X) - X. Nonlinear mixing of input vector
!
! Modified to call SVD routines
!******************************************************************************

Subroutine Anderson_Mix(AC, X, F)
Save
  Type  (Anderson_Context), Intent(INOUT) :: AC
  Real*8,                  Intent(INOUT) :: X(:)
  Real*8,                  Intent(IN)    :: F(:)

  Integer :: i, slot, currentdim , n ,j
  Real*8 :: term
  Real*8  :: tmp

  !** First determine where to store the new correction vectors ***
  AC%slot = AC%slot + 1
  If (AC%Slot>AC%Nmax) AC%Slot = 1

  If ((AC%N < 0) .OR. (AC%Nmax == 0)) then  !** Simple mixing for 1st time ***
     AC%Xprev = X
     X = X + AC%NewMix*F
  else
     slot = AC%Slot

     AC%DF(:,slot) = F - AC%Fprev   !** Make new DF vector
     AC%DX(:,slot) = X - AC%Xprev   !** Make new DX vector

     currentdim=Min(AC%N+1,AC%Nmax)
     Do i=1, currentdim              !*** Add row/col to matrix
        term = DOT_PRODUCT(AC%DF(:,i), AC%DF(:,slot))
        AC%Matrix(i,slot) = term
        if (i /= slot) AC%Matrix(slot,i) = (term)

        AC%Gamma(i) = DOT_PRODUCT(AC%DF(:,i), F)
     End Do


     AC%DupMatrix = AC%Matrix
! not needed (SVD) AC%DupMatrix(slot,slot) = (1+AC%w0) * AC%DupMatrix(slot,slot)
!     Call ZHESV('L', Slot, 1, AC%DupMatrix(1,1), AC%Nmax, &
!          AC%IPIV(1), AC%Gamma(1), AC%Nmax, AC%Work(1), AC%LWork, i)

!     Call ZGESV(Slot, 1, AC%DupMatrix(1,1), AC%Nmax, AC%IPIV(1), &
!          AC%Gamma(1), AC%Nmax, i)
!    Call ZGESV(currentdim, 1, AC%DupMatrix(1,1), AC%Nmax, AC%IPIV(1), &
!         AC%Gamma(1), AC%Nmax, i)

     n = AC%Nmax;   j= currentdim
     Call DGESDD('A',j,j,AC%DupMatrix(1,1),n,AC%S(1), &
          AC%U(1,1),n,AC%VT(1,1),n,AC%Work(1),AC%Lwork, AC%IPIV(1),i)
     If (i /= 0) then
        Write(AC%Err_Unit,*) 'Anderson_Mix: Error in DGESDD. Error=',i
        tmp = 0
        tmp = 1.d0/tmp
        STOP
     end If

     Write(AC%Err_Unit,*) 'in Anderson_Mix -- completed SVD with values'
     Write(AC%Err_Unit,'(1p5e15.7)') (AC%S(i),i=1,j)

     AC%Work(1:j) = AC%Gamma(1:j)
     AC%Gamma = 0
     tmp=MAX(ABS(AC%S(1))/AC%ConditionNo,AC%Machaccur)
      Do i=1,j
        If (ABS(AC%S(i)).gt.tmp) then
        AC%Gamma(1:j)=AC%Gamma(1:j)+&
           (AC%VT(i,1:j))*DOT_PRODUCT(AC%U(1:j,i),AC%Work(1:j))/AC%S(i)
        EndIf
      Enddo


     AC%Xprev = X

     !*** Now calculate the new vector ***
     X = X + AC%NewMix*F

    !Do i=1, Min(AC%N+1,AC%Nmax)     !*** Add row/col to matrix
     Do i=1, currentdim               ! updated vector
        X = X - AC%Gamma(i)*(AC%DX(:,i) + AC%NewMix*AC%DF(:,i))
     End Do
  End If
  

  AC%Fprev = F

  AC%N = AC%N + 1
  If (AC%N > AC%Nmax) AC%N = AC%Nmax

  Return
End Subroutine

!******************************************************************************
!
!  Anderson_ResetMix - Resets the mixing history to None
!
!     AC - Anderson context to reset
!
!******************************************************************************

Subroutine Anderson_ResetMix(AC)
  Type  (Anderson_Context), Intent(INOUT) :: AC

  AC%N = -1
  AC%Slot = -1

  Return
End Subroutine

!******************************************************************************
!
!  FreeAnderson - Frees all the data associated with the AC data structure
!    
!      AC -Pointer to the Anderson context to free
!
!******************************************************************************

Subroutine FreeAnderson(AC)
   Type (Anderson_Context), Pointer :: AC  

   DeAllocate(AC%Xprev, AC%Fprev , AC%DX, AC%DF, AC%Matrix, AC%Gamma)
   DeAllocate(AC%DupMatrix, AC%U, AC%VT, AC%Work, AC%RWork, AC%IPIV, AC%S)
   DeAllocate(AC)

   Return
end Subroutine

!******************************************************************************
!
!  InitAnderson - Initializes and Anderson_Context data structure for use
!
!   AC       - Anderson context created and returned
!   Err_Unit - Output error unit
!   Nmax     - Max number of vectors to keep
!   VecSize  - Size of each vector
!   NewMix   - Mixing factor
!
!******************************************************************************

Subroutine InitAnderson(AC, Err_Unit, Nmax, VecSize,  NewMix, CondNo)
   Type (Anderson_Context), Pointer     :: AC
   Integer,                 Intent(IN)  :: Err_Unit
   Integer,                 Intent(IN)  :: Nmax
   Integer,                 Intent(IN)  :: VecSize
   Real*8,                    Intent(IN)  :: NewMix
   Real*8,                    Intent(IN)  :: CondNo

   Integer :: i
   Real*8    :: tmp , a1,a2,a3


   Allocate(AC)   !** Allocate the pointer

   AC%Nmax = Nmax          !*** Store the contants
   AC%VecSize = VecSize
   AC%NewMix = NewMix
   AC%Err_Unit = Err_Unit

   AC%N = -1                !** Init the rest of the structure
   AC%Slot = -1
!  AC%Lwork = 2*Nmax
!  Allocate(AC%Gamma(Nmax), AC%Work(AC%Lwork), AC%Fprev(VecSize), &
!       AC%DF(VecSize, Nmax), AC%Matrix(Nmax, Nmax), AC%IPIV(Nmax), &
!       AC%DX(VecSize, Nmax), AC%Xprev(VecSize), &
!       AC%DupMatrix(Nmax, Nmax), STAT=i)

   Allocate(AC%Xprev(VecSize), AC%Fprev(VecSize) , AC%DX(VecSize,Nmax), &
            AC%DF(VecSize,Nmax), AC%Matrix(Nmax,Nmax) , AC%Gamma(Nmax), &
            Stat=i)


   If (i /= 0) then
      Write(Err_Unit,*) 'InitAnderson: Allocate Error! Error=',i
      Write(Err_Unit, *) 'InitAnderson: Nmax=',Nmax, ' * VecSize=',VecSize
      tmp = 0
      tmp = 1.d0/tmp
      STOP
   End If

   AC%Lwork=5*Nmax*Nmax+10*Nmax
   AC%LRwork= 5*Nmax*Nmax+7*Nmax
   AC%ConditionNo= CondNo
   
   ! Calculate machine accuracy
   AC%Machaccur = 0
   a1 = 4.d0/3.d0
   do while (AC%Machaccur == 0.d0)
      a2 = a1 - 1.d0
      a3 = a2 + a2 + a2
      AC%Machaccur = ABS(a3 - 1.d0)
   enddo
   
   Write(Err_Unit,*) 'Machaccur = ', AC%Machaccur
   
   Allocate(AC%DupMatrix(Nmax,Nmax), AC%U(Nmax, Nmax), AC%VT(Nmax,Nmax), &
        AC%Work(AC%Lwork), AC%RWork(AC%LRWork), AC%IPIV(8*Nmax), &
        AC%S(Nmax), STAT=i)
   
   AC%Matrix = 0

   Return
End Subroutine


End Module
